(* 
  Abstract Syntax Tree for Gwiz
  Based on MicroC 
  File: ast.ml 
  Authors: Ashley Kim, Katherine Duff
  PLT Spring 2021 
*) 

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq | And | Or

type typ = Int | String | Bool | Void | Double | Trackpoint | ArrayType of typ (*| TrackArrayType of typ * int *)| File
type bind = typ * string   

type uop = Neg | Not

type expr =
    Literal of int
  | DLit of string
  | SLit of string
  | BLit of bool
  | Id of string
  | Binop of expr * op * expr
  | Unop of uop *  expr
  | Assign of string * expr  
  | Call of string * expr list
  | TrackpointLit of expr * expr * expr 
  | TrackpointAccess of expr * string 
  | ArrayLiteral of expr list
  | ArrayAccess of expr * expr
  | ArrayAppend of expr * expr
  (*
  | TrackArrayLiteral of expr list
  | TrackArrayAccess of expr * expr
  | TrackArrayInit of expr * int *)
  | Noexpr


type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt

type func_decl = {
    typ : typ;
    fname : string;
    formals : bind list;
    locals  : bind list;
    body : stmt list;
}

type program = bind list * func_decl list


(* Pretty-printing functions. *)
let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"



let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | DLit(l) -> l
  | BLit(true) -> "true"
  | BLit(false) -> "false"
  | SLit(l) -> l 
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e 
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | TrackpointAccess(e1, s2) -> string_of_expr e1 ^ "." ^ s2 
  | TrackpointLit(e1, e2, e3) -> "(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ")"   
  | ArrayLiteral(e1) -> "[" ^ String.concat ", " (List.map string_of_expr e1) ^ "]"
  | ArrayAccess(v, e) -> string_of_expr v ^ "-<" ^ string_of_expr e ^ ">"
  | ArrayAppend(v, e) -> string_of_expr v ^ "+<" ^ string_of_expr e ^ ">"
  (*| TrackArrayLiteral (e1) -> "<[" ^ String.concat ", " (List.map string_of_expr e1) ^ "]>"
  | TrackArrayAccess(v,e) -> string_of_expr v ^ "<[" ^ string_of_expr e ^ "]>"
  | TrackArrayInit(i, s) -> "new" ^ string_of_expr i ^ "<[" ^ string_of_int s ^ "]>"  *)
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s

  let rec string_of_typ = function
    Int -> "int"
  | Double -> "double"
  | Bool -> "bool"
  | String -> "string"
  | Void -> "void"
  | Trackpoint -> "Trackpoint"
  | ArrayType(t) -> string_of_typ t ^ "[]"
 (* | TrackArrayType(t, l) -> string_of_typ t ^ "<[" ^ string_of_int l ^ "]>" *)
  | File -> "file"

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.locals) ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
