
/* Ocamlyacc parser for GPX Wizard */

%{
open Ast
let parse_error s =
      begin
        try
          let start_pos = Parsing.symbol_start_pos ()
          and end_pos = Parsing.symbol_end_pos () in
          Printf.printf "File \"%s\", line %d, characters %d-%d: \n"
            start_pos.pos_fname
            start_pos.pos_lnum
            (start_pos.pos_cnum - start_pos.pos_bol)
            (end_pos.pos_cnum - start_pos.pos_bol)
        with Invalid_argument(_) -> ()
      end;
      Printf.printf "Syntax error: %s\n" s;
      raise Parsing.Parse_error
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE LBRAC RBRAC PERIOD COMMA /*seperators*/
%token PLUS MINUS TIMES DIVIDE ASSIGN /*binary operators */
%token NOT EQ NEQ LT LEQ GT GEQ AND OR NEG/*logical operators*/
%token IF ELSE FOR RETURN INT BOOL DOUBLE STRING VOID TRACKPOINT/*keywords*/
%token <int> LITERAL
%token <bool> BLIT /*boolean literal*/
%token <string> ID SLIT
%token EOF


%nonassoc NOELSE
%nonassoc ELSE
%left SEMI
%right ASSIGN
%left PERIOD
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right NOT
%right NEG


%start program
%type <Ast.program> program

%%

program:
  decls EOF { $1 }

decls:
    /*nothing*/ {   ([], [])      }
  | decls vdecl { (($2 :: fst $1), snd $1) }
  | decls fdecl { (fst $1, ($2 :: snd $1)) }

fdecl: /* function declaration */
    typ ID LPAREN params RPAREN LBRACE vdecl_list stmt_list RBRACE {
    { typ = $1; fname = $2; formals = List.rev $4; locals = List.rev $7;
    body = List.rev $8 } }

params: /*function parameters*/
    /*nothing*/ { [] }
  | params_list { $1 }

params_list:
    typ ID { [($1, $2)] }
  | params_list COMMA typ ID { ($3, $4) :: $1 }

typ:
    INT { Int }
  | DOUBLE { Double }
  | BOOL { Bool }
  | VOID { Void }
  | STRING { String }
  | TRACKPOINT { Double, Double, DateTime }

vdecl_list:
    /*nothing*/ { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl: /*variable declaration */
  typ ID SEMI { ($1, $2) }

stmt_list:
    /*nothing*/ { [] }
  | stmt_list stmt { $2 :: $1 }

stmt: /*statements*/
    expr SEMI { Expr ($1) }
  | RETURN expr_opt SEMI { Return ($2)}
  | LBRACE stmt_list RBRACE { Block(List.rev $2)} /*nesting statements*/
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt { If($3, $5, $7)}
  | FOR LPAREN expr SEMI expr SEMI expr_opt RPAREN stmt { For($3, $5, $7, $9) }
  | WHILE LPAREN expr RPAREN stmt           { While($3, $5)         }

expr_opt: /*when an expression is optional*/
    /*nothing*/ { Noexpr }
  | expr { $1 }

expr:
    LITERAL { Literal($1) }
  | BLIT { Blit($1) }
  | ID { Id($1) }
  | SLIT { Slit($1) }
  | expr PLUS expr { Binop($1, Add, $3) }
  | expr MINUS expr { Binop($1, Sub, $3) }
  | expr TIMES expr { Binop($1, Mult, $3) }
  | expr DIVIDE expr { Binop($1, Div, $3) }
  | expr EQ expr { Binop($1, Equal, $3) }
  | expr NEQ expr { Binop($1, Neq, $3) }
  | expr LT expr { Binop($1, Less, $3) }
  | expr LEQ expr { Binop($1, Leq, $3) }
  | expr GT expr { Binop($1, Greater, $3) }
  | expr GEQ expr { Binop($1, Geq, $3) }
  | expr AND expr { Binop($1, And, $3) }
  | expr OR expr { Binop($1, Or, $3) }
  | NEG expr %prec NOT{ Unop(Neg, $2) }
  | NOT expr { Unop(Not, $2) }
  | ID ASSIGN expr { Assign($1, $3) }
  | ID LPAREN args_opt RPAREN { Call($1, $3) } /*function calling*/


args_opt:
    /*nothing*/ { [] }
  | args_list { List.rev $1 }

args_list:
    expr { [$1] }
  | args_list COMMA expr { $3 :: $1 }
